% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcess.R
\name{projectInputs}
\alias{projectInputs}
\alias{projectInputs.default}
\alias{projectInputs.Raster}
\alias{projectInputs.SpatVector}
\alias{projectInputs.SpatRaster}
\alias{projectInputs.sf}
\alias{projectInputs.Spatial}
\title{Project \code{Raster*} or \code{Spatial*} or \code{sf} objects}
\usage{
projectInputs(
  x,
  targetCRS,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\method{projectInputs}{default}(x, targetCRS, ...)

\method{projectInputs}{Raster}(
  x,
  targetCRS = NULL,
  verbose = getOption("reproducible.verbose", 1),
  rasterToMatch = NULL,
  cores = NULL,
  useGDAL = getOption("reproducible.useGDAL", TRUE),
  ...
)

\method{projectInputs}{SpatVector}(
  x,
  targetCRS,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\method{projectInputs}{SpatRaster}(
  x,
  targetCRS = NULL,
  verbose = getOption("reproducible.verbose", 1),
  rasterToMatch = NULL,
  cores = NULL,
  useGDAL = getOption("reproducible.useGDAL", TRUE),
  ...
)

\method{projectInputs}{sf}(
  x,
  targetCRS,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\method{projectInputs}{Spatial}(
  x,
  targetCRS,
  verbose = getOption("reproducible.verbose", 1),
  ...
)
}
\arguments{
\item{x}{A \code{Raster*}, \code{Spatial*} or \code{sf} object}

\item{targetCRS}{The CRS of x at the end  of this function (i.e., the goal)}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \code{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Passed to \code{\link[raster]{projectRaster}}.}

\item{rasterToMatch}{Template \code{Raster*} object passed to the \code{to} argument of
\code{\link[raster]{projectRaster}}, thus will changing the
resolution and projection of \code{x}.
See details in \code{\link{postProcess}}.}

\item{cores}{An \code{integer*} or \code{'AUTO'}. This will be used if \code{gdalwarp} is
triggered. \code{'AUTO'*} will calculate 90% of the total
number of cores in the system, while an integer or rounded
float will be passed as the exact number of cores to be used.}

\item{useGDAL}{Logical or \code{"force"}.
Defaults to \code{getOption("reproducible.useGDAL" = TRUE)}.
If \code{TRUE}, then this function will use \code{gdalwarp} only when not
small enough to fit in memory (i.e., \emph{if the operation fails} the
\code{raster::canProcessInMemory(x, 3)} test). Using \code{gdalwarp} will
usually be faster than \code{raster::projectRaster}, the function used
if this is \code{FALSE}. Since since the two options use different algorithms,
there may be different projection results. \code{"force"} will cause it to
use GDAL regardless of the memory test described here.}
}
\value{
A file of the same type as starting, but with projection (and possibly
other characteristics, including resolution, origin, extent if changed).
}
\description{
A simple wrapper around the various different tools for these GIS types.
}
\examples{
# Add a study area to Crop and Mask to
# Create a "study area"
library(sp)
library(raster)
ow <- setwd(tempdir())

# make a SpatialPolygon
coords1 <- structure(c(-123.98, -117.1, -80.2, -100, -123.98, 60.9, 67.73, 65.58, 51.79, 60.9),
                     .Dim = c(5L, 2L))
Sr1 <- Polygon(coords1)
Srs1 <- Polygons(list(Sr1), "s1")
shpEcozone <- SpatialPolygons(list(Srs1), 1L)
crs(shpEcozone) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"

# make a "study area" that is subset of larger dataset
coords <- structure(c(-118.98, -116.1, -99.2, -106, -118.98, 59.9, 65.73, 63.58, 54.79, 59.9),
                    .Dim = c(5L, 2L))
Sr1 <- Polygon(coords)
Srs1 <- Polygons(list(Sr1), "s1")
StudyArea <- SpatialPolygons(list(Srs1), 1L)
crs(StudyArea) <- crs(shpEcozone)
projString <- "+proj=utm +zone=15 +ellps=GRS80 +datum=NAD83 +units=m +no_defs"
StudyArea <- sp::spTransform(StudyArea, CRSobj = projString)

##########
shpEcozonePostProcessed <- postProcess(shpEcozone, studyArea = StudyArea)


# Try manually, individual pieces
shpEcozoneReprojected <- projectInputs(shpEcozone, StudyArea)
shpEcozoneCropped <- cropInputs(shpEcozone, StudyArea)
shpEcozoneClean <- fixErrors(shpEcozone)
shpEcozoneMasked <- maskInputs(shpEcozone, StudyArea)

# With terra
if (require("terra")) {
  opts <- options("reproducible.useTerra" = TRUE)
  vectEcozone <- terra::vect(sf::st_as_sf(shpEcozone)) # direct conversion throws warning

  # If input is Spatial object --> return will also be Spatial
  shpEcozonePostProcessed <- postProcess(shpEcozone, studyArea = StudyArea)
  # Try manually, individual pieces -- Note functions are different
  shpEcozoneReprojected <- projectInputs(shpEcozone, StudyArea)
  shpEcozoneMasked <- maskInputs(shpEcozone, StudyArea)
  shpEcozoneCropped <- cropInputs(shpEcozone, StudyArea)

  # If input is Spat object --> return will also be Spat
  vectEcozonePostProcessed <- postProcess(vectEcozone, studyArea = StudyArea)
  # Try manually, individual pieces -- Note functions are different
  vectEcozoneMasked <- maskInputs(vectEcozone, StudyArea)
  VectEcozoneReprojected <- projectInputs(vectEcozone, StudyArea)
  vectEcozoneCropped <- cropInputs(vectEcozone, StudyArea)



  # Note these two have different function names --> methods for cropInputs and fixErrors
  #    are not implemented yet
  shpEcozoneClean <- fixErrorsTerra(vectEcozone)

  options(opts)
}

setwd(ow)
}
